
        # TODO for Multi:
        # - work on non-capture calls, for temporary backwards compatibility
        # - dispatch on types,
        # - dispatch on named args
        # - integrate with the Grammar
        # - how to make it work with same-name multis on outer Pads

        # @($.long_names)  - subroutine list
        #    .signature
        #        .invocant
        #        .array
        #        .hash

use v6-alpha;
class Multi is Code {
    has @.long_names;    # sub/method multi-dispatch

    method add_variant ( $code ) {
        if !( defined( self.long_names ) ) {
            self.long_names = [ ];   # XXX accessor init bug
        };

        (self.long_names).push( $code );
    }


    method select ( $capture ) {
        if !( defined( self.long_names ) ) {
            die "can't resolve Multi dispatch (the Multi has no long names) ";
        };

        my @candidates = ( );

        for @(self.long_names) -> $sub {
            if (($sub.signature).arity == $capture.arity) {
                @candidates.push( $sub );
            };
        };

        if @candidates.elems == 1 {
            return @candidates[0];
        };
        if @candidates.elems > 1 {
            die "Multi dispatch is ambiguous. Matching variants "~(@candidates).perl;
        }

        die "can't resolve Multi dispatch - no matching variant found";
    }

    method perl {
        'Multi.new()'
    }
}

class MultiToken is Code {
    has %.token_length;  # token:sym  multi-dispatch
    method add_token_variant ( $code, $sym ) {

        # XXX - fixme
        # http://moritz.faui2k3.org/irclog/out.pl?channel=perl6;date=2007-09-17#i_104660
        # <TimToady> sym is a way of specifying the identifier of some tokens and later some operator names,
        #   but it's not the definition of longest token
        #   if the rule says "<sym> 'bar'" then the token would be $bar (at least)
        #   it's basically everything DFAable up to <ws> or ::

        my $len = $sym.chars;
        #say "add_token_variant: $sym : $len ";
        if !( defined( self.token_length ) ) {
            self.token_length = { };   # XXX accessor init bug
        };
        if !( defined( ( self.token_length ){$len} ) ) {
            ( self.token_length ){$len} = { };
        };
        ( ( self.token_length ){$len} ){ $sym } = Multi.new();
        ( ( ( self.token_length ){$len} ){$sym} ).add_variant( $code );
    }
    method select ( $capture ) {
        unless $capture.isa( 'Capture' ) {
            die "the parameter to Multi.select must be a Capture";
        };

        if     defined( self.token_length )
            && ( self.token_length ).keys
        {
            my @len = ( ( self.token_length ).keys );
            #say "lengths : @len[] ";
            @len = @len.sort( sub { @_[1] <=> @_[0] } );
            #say "lengths : @len[] ";
            #say "string: $_";
            #say "string len: ",$_.chars;

            for @len -> $len {
                if ( $_.chars ) >= $len {
                    my $s = substr( $_, 0, $len );   # XXX - pos?
                    #say "# len: $len - $s";
                    my %syms = ( self.token_length ){$len};
                    #say "# syms: ", ( %syms.keys );

                    for %syms.keys -> $sym {
                        if $s eq $sym {
                            #say "found!";
                            return ( %syms{$sym} ).select( $capture );
                        }
                    }
                };
            };
        };
    }
    method perl {
        'MultiToken.new()'
    }
}

=begin

=head1 AUTHORS

The Pugs Team E<lt>perl6-compiler@perl.orgE<gt>.

=head1 SEE ALSO

The Perl 6 homepage at L<http://dev.perl.org/perl6>.

The Pugs homepage at L<http://pugscode.org/>.

=head1 COPYRIGHT

Copyright 2007 by Flavio Soibelmann Glock and others.

This program is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

See L<http://www.perl.com/perl/misc/Artistic.html>

=end
